/*
 *	Filesystem Map List / Diff 1.0
 *
 *	(c) 1997 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include "types.h"
#include "fastbuf.h"
#include "filemap.h"

#define MAXDEPTH 32

struct df {
  struct fastbuf *f;
  byte path[MAX_NAME_SIZE], param[MAX_NAME_SIZE];
  uns depth;
  int eof;
  byte *pp[MAXDEPTH];
  int type, tych;
  ulg perm, uid, gid, mod;
};

static struct df *in, *cc;

static struct df *
df_open(byte *name)
{
  struct df *k = xmalloc(sizeof(struct df));
  byte buf[4];

  k->f = bopen(name, O_RDONLY, 65536);
  k->path[0] = 0;
  k->depth = k->eof = 0;
  k->pp[0] = k->path;
  bread(k->f, buf, 4);
  if (memcmp(buf, FM_IDENT, 4) || bgetl(k->f) != FM_VERSION)
    die("%s: Invalid format", name);
  return k;
}

static int
df_read(struct df *d)
{
  struct fastbuf *f = d->f;
  byte *out;
  int l;

again:
  d->type = bgetc(f);
  if (d->type < 0)
    {
      if (d->depth)
	die("%s: Premature EOF", f->name);
      d->eof = 1;
      return 0;
    }
  if (d->type == E_EOD)
    {
      if (!d->depth)
	die("%s: Hollow tree", f->name);
      d->depth--;
      goto again;
    }
  l = bgetc(f);
  bread(f, out = d->pp[d->depth], l);
  out[l] =  0;
  d->perm = bgetw(f);
  d->uid = bgetw(f);
  d->gid = bgetw(f);
  d->mod = bgetl(f);
  out = d->param;
  out[0] = 0;
  switch (d->type)
    {
    case E_FILE:
      sprintf(out, "%d", bgetl(f));
      break;
    case E_FILE_CRC:
      out += sprintf(out, "%d", bgetl(f));
      sprintf(out, " [%08x]", bgetl(f));
      break;
    case E_SYMLINK:
    case E_HARDLINK:
      l = bgetc(f);
      out += sprintf(out, "-> ");
      bread(f, out, l);
      out[l] = 0;
      break;
    case E_CHAR:
    case E_BLOCK:
      l = bgetl(f);
      sprintf(out, "%d/%d", l / 256, l % 256);
      break;
    case E_DIR:
      if (d->depth >= MAXDEPTH)
	die("%s: Maxdepth exceeded\n", f->name);
      out = d->pp[++d->depth] = d->path + strlen(d->path) + 1;
      out[-1] = '/';
      out[0] = 0;
      break;
    case E_FIFO:
    case E_SOCK:
      break;
    default:
      die("%s: Invalid type %02x", f->name, d->type);
    }
  d->tych = "EDFFSHCBfs" [d->type];
  return 1;
}

static int
df_cmp(struct df *a, struct df *b)
{
  byte *x, *y;
  uns xx, yy;

  if (a->eof)
    return !b->eof;
  if (b->eof)
    return -1;

  x = a->path;
  y = b->path;
  while (*x && *y)
    {
      xx = *x++;
      yy = *y++;
      if (xx == '/')
	xx = 0;
      if (yy == '/')
	yy = 0;
      if (xx < yy)
	return -1;
      if (xx > yy)
	return 1;
    }
  if (*x)
    return 1;
  if (*y)
    return -1;
  return 0;
}

static void
df_print(struct df *d, int tag)
{
  if (tag)
    {
      putchar(tag);
      putchar(' ');
    }
  printf("%c %s P%3o U%d G%d M%u%s%s\n",
	 d->tych,
	 d->path,
	 d->perm,
	 d->uid,
	 d->gid,
	 d->mod,
	 d->param[0] ? " " : "",
	 d->param);
}

static void
comp(void)
{
  df_read(in);
  df_read(cc);
  while (!in->eof || !cc->eof)
    switch (df_cmp(in, cc))
      {
      case -1:
	df_print(in, '-');
	df_read(in);
	break;
      case 1:
	df_print(cc, '+');
	df_read(cc);
	break;
      default:
	if (   in->type != cc->type
	    || in->perm != cc->perm
	    || in->uid != cc->uid
	    || in->gid != cc->gid
	    || (in->mod != cc->mod && in->type != E_DIR)
	    || strcmp(in->param, cc->param))
	  {
	    df_print(in, '?');
	    df_print(cc, '!');
	  }
	df_read(in);
	df_read(cc);
      }
}

static void
show(void)
{
  while (df_read(in))
    df_print(in, 0);
}

int
main(int argc, char **argv)
{
  if (argc != 2 && argc != 3)
    {
      fprintf(stderr, "Usage: fdiff <map>\t\t\tList a map\n"
	              "       fdiff <map1> <map2>\t\tFind differences between two maps\n");
      return 1;
    }
  in = df_open(argv[1]);
  if (argc == 2)
    show();
  else
    {
      cc = df_open(argv[2]);
      comp();
    }
  return 0;
}
