/*
 *	Filesystem Scanner 1.0
 *
 *	(c) 1997 Martin Mares <mj@atrey.karlin.mff.cuni.cz>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <dirent.h>
#include <alloca.h>
#include <unistd.h>

#include "types.h"
#include "fastbuf.h"
#include "filemap.h"
#include "md5.h"

static struct fastbuf *out;
static int do_crc;
static byte dirbuf[MAX_NAME_SIZE], linkbuf[MAX_NAME_SIZE];

static void do_scan(byte *sep, byte *segstart);

struct de {
  struct de *next;
  byte name[1];
};

static int
cmp(const struct de **x, const struct de **y)
{
  const struct de *X = *x;
  const struct de *Y = *y;

  return strcmp(X->name, Y->name);
}

static void
traverse_dir(byte *sep)
{
  DIR *d;
  struct dirent *e;
  struct de *x, *first, **last = &first, **idx;
  int cnt, i;

  d = opendir(dirbuf);
  *sep++ = '/';
  *sep = 0;
  if (!d)
    {
      fprintf(stderr, "opendir(%s): %m", dirbuf);
      return;
    }
  cnt = 0;
  while (e = readdir(d))
    {
      if (!strcmp(e->d_name, ".") || !strcmp(e->d_name, ".."))
	continue;
      x = alloca(sizeof(struct de) + strlen(e->d_name));
      *last = x;
      last = &x->next;
      strcpy(x->name, e->d_name);
      cnt++;
    }
  closedir(d);
  *last = NULL;

  if (cnt)
    {
      idx = alloca(sizeof(struct de *) * cnt);
      i = 0;
      for(x=first; x; x=x->next)
	idx[i++] = x;
      qsort(idx, cnt, sizeof(struct de *), (int (*)(const void *, const void *)) cmp);
      for(i=0; i<cnt; i++)
	{
	  x = idx[i];
	  strcpy(sep, x->name);
	  do_scan(sep + strlen(x->name), sep);
	}
    }
  bputc(out, E_EOD);
}

static ulg
file_crc(void)
{
  byte fbuf[16384];
  int fd, c;
  struct MD5Context md;
  ulg dig[4];

  MD5Init(&md);
  fd = open(dirbuf, O_RDONLY);
  if (fd < 0)
    {
      fprintf(stderr, "open(%s): %m\n", dirbuf);
      return 0;
    }
  while ((c = read(fd, fbuf, sizeof(fbuf))) > 0)
    MD5Update(&md, fbuf, c);
  if (c < 0)
    fprintf(stderr, "read(%s): %m\n", dirbuf);
  close(fd);
  MD5Final((byte *) dig, &md);
  return dig[0] ^ dig[1] ^ dig[2] ^ dig[3];
}

#define HASH_TABLE_SIZE 4099		/* Prime */

struct lent {
  struct lent *next;
  ulg rdev, ino;
  byte name[1];
};

static struct lent *lenthash[HASH_TABLE_SIZE];

static uns
is_hard_link(byte *name, ulg rdev, ulg ino)
{
  uns hv = (rdev ^ (rdev << 17) ^ ino) % HASH_TABLE_SIZE;
  struct lent *l;

  for(l=lenthash[hv]; l; l=l->next)
    {
      if (l->rdev == rdev && l->ino == ino)
	{
	  strcpy(linkbuf, l->name);
	  return 1;
	}
    }
  l = xmalloc(sizeof(struct lent) + strlen(name));
  l->next = lenthash[hv];
  lenthash[hv] = l;
  l->rdev = rdev;
  l->ino = ino;
  strcpy(l->name, name);
  return 0;
}

static void
do_scan(byte *sep, byte *segstart)
{
  struct stat st;
  int type, i;

  *sep = 0;
  if (!dirbuf[0])
    strcpy(dirbuf, "/");
  if (lstat(dirbuf, &st) < 0)
    {
      fprintf(stderr, "%s: %m\n", dirbuf);
      return;
    }
  switch (st.st_mode & S_IFMT)
    {
    case S_IFSOCK:
      type = E_SOCK;
      break;
    case S_IFLNK:
      type = E_SYMLINK;
      i = readlink(dirbuf, linkbuf, sizeof(linkbuf) - 1);
      if (i < 0)
	{
	  fprintf(stderr, "readlink(%s): %m\n", dirbuf);
	  return;
	}
      linkbuf[i] = 0;
      break;
    case S_IFREG:
      type = do_crc ? E_FILE_CRC : E_FILE;
      break;
    case S_IFBLK:
      type = E_BLOCK;
      break;
    case S_IFDIR:
      type = E_DIR;
      break;
    case S_IFCHR:
      type = E_CHAR;
      break;
    case S_IFIFO:
      type = E_FIFO;
      break;
    default:
      fprintf(stderr, "%s: Unknown file type\n", dirbuf);
      return;
    }
  if (type != E_DIR && st.st_nlink > 1 && is_hard_link(dirbuf, st.st_rdev, st.st_ino))
    type = E_HARDLINK;
  bputc(out, type);
  bputc(out, sep - segstart);
  bwrite(out, segstart, sep - segstart);
  bputw(out, st.st_mode & 07777);
  bputw(out, st.st_uid);
  bputw(out, st.st_gid);
  bputl(out, (st.st_ctime > st.st_mtime) ? st.st_ctime : st.st_mtime);
  switch (type)
    {
    case E_DIR:
      traverse_dir(sep);
      break;
    case E_FILE:
      bputl(out, st.st_size);
      break;
    case E_FILE_CRC:
      bputl(out, st.st_size);
      bputl(out, file_crc());
      break;
    case E_SYMLINK:
    case E_HARDLINK:
      i = strlen(linkbuf);
      if (i > 255)
	{
	  fprintf(stderr, "%s: Link destination too long!\n", dirbuf);
	  strcpy(linkbuf, "???");
	  i = 3;
	}
      bputc(out, i);
      bwrite(out, linkbuf, i);
      break;
    case E_CHAR:
    case E_BLOCK:
      bputl(out, st.st_rdev);
      break;
    case E_FIFO:
    case E_SOCK:
      break;
    default:
      fprintf(stderr, "Something is wrong!\n");
      exit(23);
    }
}

static void
scan(byte *path)
{
  byte *k;

  if (!*path)
    path = ".";
  strcpy(dirbuf, path);
  k = dirbuf + strlen(dirbuf) - 1;
  if (*k != '/')
    k++;
  do_scan(k, dirbuf);
}

int
main(int argc, char **argv)
{
  int i;

  if (argc < 3)
    {
      fprintf(stderr, "Usage: fscan outfile [+c] <list-of-dirs>\n");
      return 1;
    }
  out = bopen(argv[1], O_CREAT | O_TRUNC | O_WRONLY, 65536);
  bwrite_slow(out, FM_IDENT, 4);
  bputl(out, FM_VERSION);
  for(i=2; i<argc; i++)
    {
      byte *k = argv[i];
      if ((k[0] == '-' || k[0] == '+') && strlen(k) == 2)
	{
	  int val = (k[0] == '+');
	  switch (k[1])
	    {
	    case 'c':
	      do_crc = val;
	      break;
	    default:
	      fprintf(stderr, "Unknown switch: %s", k);
	    }
	}
      else scan(argv[i]);
    }
  bclose(out);
  return 0;
}
